#define CURL_STATICLIB
#include <curl/curl.h>

#define _RETAIL
#include "GMLuaModule.h"
#include "gm_curl.h"

CURLM *multi_handle;

int num_curl_handles = 0;

struct MemoryStruct {
	char *memory;
	size_t size;
	int reference;
	curl_httppost *post;
};

static size_t WriteMemoryCallback(void *ptr, size_t size, size_t nmemb, void *data)
{
	size_t realsize = size * nmemb;
	struct MemoryStruct *mem = (struct MemoryStruct *)data;

	mem->memory = (char *)realloc(mem->memory, mem->size + realsize + 1);

	if (mem->memory) {
		memcpy(&(mem->memory[mem->size]), ptr, realsize);
		mem->size += realsize;
		mem->memory[mem->size] = 0;
	}

	return realsize;
}

LUA_FUNCTION(curl_poll)
{
	ILuaInterface *gLua = Lua();

	int num_handles;

	while(CURLM_CALL_MULTI_PERFORM ==
		curl_multi_perform(multi_handle, &num_handles));

	for(;;)
	{
		int msgs;
		CURLMsg *msg;

		msg = curl_multi_info_read(multi_handle, &msgs);

		if(msg == NULL || msg->msg == CURLMSG_NONE)
			break;

		if(msg->msg == CURLMSG_DONE)
		{
			MemoryStruct *chunk = NULL;
			curl_easy_getinfo(msg->easy_handle, CURLINFO_PRIVATE, &chunk);

			gLua->PushReference(chunk->reference);

			if(msg->data.result == CURLE_OK)
				gLua->Push(chunk->memory);
			else
				gLua->Push(false);

			gLua->Call(1, 0);

			CURL *handle = msg->easy_handle;

			curl_multi_remove_handle(multi_handle, handle);

			curl_easy_cleanup(handle);

			if(chunk->post)
				curl_formfree(chunk->post);

			--num_curl_handles;

			gLua->FreeReference(chunk->reference);

			free(chunk->memory);
			free(chunk);
		}
	}
	return 0;
}

CURL *GetNewHandle(ILuaInterface *gLua)
{
	gLua->CheckType(1, GLua::TYPE_STRING);
	gLua->CheckType(2, GLua::TYPE_FUNCTION);

	CURL *handle = curl_easy_init();

	curl_easy_setopt(handle, CURLOPT_URL, gLua->GetString(1));
	curl_easy_setopt(handle, CURLOPT_USERAGENT, "Gmode/11.0");

	MemoryStruct *chunk = new MemoryStruct;

	chunk->memory = NULL;
	chunk->size = 0;
	chunk->post = NULL;
	chunk->reference = gLua->GetReference(2);

	curl_easy_setopt(handle, CURLOPT_WRITEFUNCTION, WriteMemoryCallback);
	curl_easy_setopt(handle, CURLOPT_WRITEDATA, (void *)chunk);

	curl_easy_setopt(handle, CURLOPT_PRIVATE, (char *)chunk);

	curl_easy_setopt(handle, CURLOPT_ENCODING, "");

	return handle;
}

void DispatchHandle(CURL *handle)
{
	curl_multi_add_handle(multi_handle, handle);
	++num_curl_handles;
}

LUA_FUNCTION(curl_get)
{
	ILuaInterface *gLua = Lua();
	
	CURL *handle = GetNewHandle(gLua);

	DispatchHandle(handle);
	return 0;
}

LUA_FUNCTION(curl_post)
{
	ILuaInterface *gLua = Lua();

	gLua->CheckType(3, GLua::TYPE_TABLE);

	CURL *handle = GetNewHandle(gLua);

	struct curl_httppost *formpost = NULL;
	struct curl_httppost *lastptr = NULL;

	ILuaObject *table = gLua->GetObject(3);

	CUtlLuaVector *tablemembers = table->GetMembers();

	for (int i = 0; i < tablemembers->Size(); i++ )
	{
		LuaKeyValue kv = (*tablemembers)[i];
	
		curl_formadd(&formpost,
			&lastptr,
			CURLFORM_COPYNAME, kv.pKey->GetString(),
			CURLFORM_COPYCONTENTS, kv.pValue->GetString(),
			CURLFORM_END);
	}

	MemoryStruct *chunk = NULL;
	curl_easy_getinfo(handle, CURLINFO_PRIVATE, &chunk);
	chunk->post = formpost;

	curl_easy_setopt(handle, CURLOPT_HTTPPOST, formpost);

	DispatchHandle(handle);

	return 0;
}

int Start( lua_State* L )
{
	ILuaInterface *gLua = Lua();

	curl_global_init(CURL_GLOBAL_ALL);

	multi_handle = curl_multi_init();

	gLua->SetGlobal("curl_get", curl_get);
	gLua->SetGlobal("curl_post", curl_post);

	ILuaObject *hookt = gLua->GetGlobal("hook");
	ILuaObject *addf = hookt->GetMember("Add");
	addf->Push();
	gLua->Push("Think");
	gLua->Push("cURLCallback");
	gLua->Push(curl_poll);
	gLua->Call(3);

	gLua->Msg("libcurl loaded: %s\n", curl_version());
	return 0;
}

int Close( lua_State* L )
{
	ILuaInterface *gLua = Lua();

	if(num_curl_handles > 0)
	{
		gLua->Msg("Blocking exit until curl queries are completed (%d).\n", num_curl_handles);
	}

	while(num_curl_handles > 0)
	{
		curl_poll(L);
		Sleep(100);
	}

	curl_multi_cleanup(multi_handle);

	curl_global_cleanup();
	return 0;
}